/* func.c   by Michael Thorpe   2005-08-12 */

#include <string.h>
#include "lisp.h"

/* This function does not incref/decref anything; returns 0 on success */
static int bindlist(scope *s,obj *names,obj *values) {
	obj *tmp;

	if(names==&nilobj) {
		if(values==&nilobj)
			return(0);
		return(-1);
	}
	if(names->type==&cons_objtype) {
		if(values->type != &cons_objtype)
			return(-1);
		if(bindlist(s,CAR(names),CAR(values)))
			return(-1);
		return(bindlist(s,CDR(names),CDR(values)));
	}
	tmp=define_local(incscoperef(s),incref(names),incref(values));
	if(tmp) {
		decref(tmp);
		return(0);
	}
	return(-1);
}

static obj *applyfunc(scope *s,obj *func,obj *args) {
	obj *prog;

	decscoperef(s);
	s=newscope(incscoperef(globalscope));
	if(!s) {
		decref(func);
		decref(args);
		return(0);
	}
	if(bindlist(s,func->value.func.args,args)) {
		decscoperef(s);
		decref(args);
		return(throwargumenterror(func));
	}
	prog=incref(func->value.func.prog);
	decref(func);
	decref(args);
	return(evaluate(s,prog));
}

static obj *func_tostring(obj *in) {
	obj *args,*func;
	obj *out=0;

	args=(in->value.func.args->type->tostring)(incref(in->value.func.args));
	if(args) {
		func=(in->value.func.prog->type->tostring)(incref(in->value.func.prog));
		if(func) {
			out=real_new_string(0,8+args->value.s.len+1+func->value.s.len+1);
			if(out) {
				memcpy(out->value.s.data,"(lambda ",8);
				memcpy(out->value.s.data+8,args->value.s.data,args->value.s.len);
				out->value.s.data[8+args->value.s.len]=' ';
				memcpy(out->value.s.data+8+args->value.s.len+1,func->value.s.data,func->value.s.len);
				out->value.s.data[8+args->value.s.len+1+func->value.s.len]=')';
			}
			decref(func);
		}
		decref(args);
	}
	decref(in);
	return(out);
}

static void drop_func(obj *o) {
	decref(o->value.func.args);
	decref(o->value.func.prog);
}

static obj *args_of_func(obj *o) {
	obj *args;

	if(&func_objtype != o->type)
		throwtypeerror(o);
	args=incref(o->value.func.args);
	decref(o);
	return(args);
}

static obj *prog_of_func(obj *o) {
	obj *prog;

	if(&func_objtype != o->type)
		throwtypeerror(o);
	prog=incref(o->value.func.prog);
	decref(o);
	return(prog);
}

static int func_eqv(obj *o1,obj *o2) {
	return(real_eqv(o1->value.func.args,o2->value.func.args) && real_eqv(o1->value.func.prog,o2->value.func.prog));
}

int func_init() {
	if(store_builtin_ns1("args-of-func",args_of_func)
	|| store_builtin_ns1("prog-of-func",prog_of_func))
		return(1);
	return(0);
}

objtype func_objtype={
	name: "func",
	tostring: func_tostring,
	dropobj: drop_func,
	applyfunc: applyfunc,
	eqv: func_eqv
};
